# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os.path

from collections import OrderedDict, defaultdict

from .metadata import read_model_metadata
from .auth import NotAuthorized
from .common import read_json_file
from .errors import ConfigurationError, ArgumentError, CubesError
from .logging import get_logger
from .calendar import Calendar
from .namespace import Namespace
from .providers import find_dimension
from .localization import LocalizationContext
from .compat import ConfigParser
from . import ext
from . import compat

__all__ = [
    "Workspace",
]


SLICER_INFO_KEYS = (
    "name",
    "label",
    "description",  # Workspace model description
    "copyright",    # Copyright for the data
    "license",      # Data license
    "maintainer",   # Name (and maybe contact) of data maintainer
    "contributors", # List of contributors
    "visualizers",  # List of dicts with url and label of server's visualizers
    "keywords",     # List of keywords describing server's cubes
    "related"       # List of dicts with related servers
)

def interpret_config_value(value):
    if value is None:
        return value
    if isinstance(value, compat.string_type):
        if value.lower() in ('yes', 'true', 'on'):
            return True
        elif value.lower() in ('no', 'false', 'off'):
            return False
    return value


def config_items_to_dict(items):
    return dict([ (k, interpret_config_value(v)) for (k, v) in items ])


class Workspace(object):
    def __init__(self, config=None, stores=None, load_base_model=True,
                 **_options):
        """Creates a workspace. `config` should be a `ConfigParser` or a
        path to a config file. `stores` should be a dictionary of store
        configurations, a `ConfigParser` or a path to a ``stores.ini`` file.

        Properties:

        * `stores` – dictionary of stores
        * `store_infos` – dictionary of store configurations
        * `namespace` – default namespace
        * `logger` – workspace logegr
        * `rot_dir` – root directory where all relative paths are looked for
        * `models_dir` – directory with models (if relative, then relative to
          the root directory)

        * `info` – info dictionary from the info file or info section
        * `calendar` – calendar object providing date and time functions
        * `ns_languages` – dictionary where keys are namespaces and values
          are language to translation path mappings.
        """

        # FIXME: **_options is temporary solution/workaround before we get
        # better configuration. Used internally. Don't use!

        # Expect to get ConfigParser instance
        if config is not None and not isinstance(config, ConfigParser):
            raise ConfigurationError("config should be a ConfigParser instance,"
                                     " but is %r" % (type(config),))
        
        if not config:
            # Read ./slicer.ini
            config = ConfigParser()

        self.store_infos = {}
        self.stores = {}

        # Logging
        # =======
        #Log to file or console
        if config.has_option("workspace", "log"):
            self.logger = get_logger(path=config.get("workspace", "log"))
        else:
            self.logger = get_logger()

        #Change to log level if necessary
        if config.has_option("workspace", "log_level"):
            level = config.get("workspace", "log_level").upper()
            self.logger.setLevel(level)

        # Set the default models path
        if config.has_option("workspace", "root_directory"):
            self.root_dir = config.get("workspace", "root_directory")
        elif "cubes_root" in _options:
            # FIXME: this is quick workaround, see note at the beginning of
            # this method
            self.root_dir = _options["cubes_root"]
        else:
            self.root_dir = ""

        if config.has_option("workspace", "models_directory"):
            self.models_dir = config.get("workspace", "models_directory")
        elif config.has_option("workspace", "models_path"):
            self.models_dir = config.get("workspace", "models_path")
        else:
            self.models_dir = ""

        if self.root_dir and not os.path.isabs(self.models_dir):
            self.models_dir = os.path.join(self.root_dir, self.models_dir)

        if self.models_dir:
            self.logger.debug("Models root: %s" % self.models_dir)
        else:
            self.logger.debug("Models root set to current directory")

        # Namespaces and Model Objects
        # ============================

        self.namespace = Namespace()

        # Cache of created global objects
        self._cubes = {}
        # Note: providers are responsible for their own caching

        # Info
        # ====

        self.info = OrderedDict()

        if config.has_option("workspace", "info_file"):
            path = config.get("workspace", "info_file")

            if self.root_dir and not os.path.isabs(path):
                path = os.path.join(self.root_dir, path)

            info = read_json_file(path, "Slicer info")
            for key in SLICER_INFO_KEYS:
                self.info[key] = info.get(key)

        elif config.has_section("info"):
            info = dict(config.items("info"))
            if "visualizer" in info:
                info["visualizers"] = [{
                    "label": info.get("label", info.get("name", "Default")),
                    "url": info["visualizer"]
                }]

            for key in SLICER_INFO_KEYS:
                self.info[key] = info.get(key)

        # Register stores from external stores.ini file or a dictionary
        if not stores and config.has_option("workspace", "stores_file"):
            stores = config.get("workspace", "stores_file")

            # Prepend the root directory if stores is relative
            if self.root_dir and not os.path.isabs(stores):
                stores = os.path.join(self.root_dir, stores)

        if isinstance(stores, compat.string_type):
            store_config = ConfigParser()
            try:
                store_config.read(stores)
            except Exception as e:
                raise ConfigurationError("Unable to read stores from %s. "
                                         "Reason: %s" % (stores, str(e) ))

            for store in store_config.sections():
                self._register_store_dict(store,
                                          dict(store_config.items(store)))

        elif isinstance(stores, dict):
            for name, store in stores.items():
                self._register_store_dict(name, store)

        elif stores is not None:
            raise ConfigurationError("Unknown stores description object: %s" %
                                                    (type(stores)))

        # Calendar
        # ========

        if config.has_option("workspace", "timezone"):
            timezone = config.get("workspace", "timezone")
        else:
            timezone = None

        if config.has_option("workspace", "first_weekday"):
            first_weekday = config.get("workspace", "first_weekday")
        else:
            first_weekday = 0

        self.logger.debug("Workspace calendar timezone: %s first week day: %s"
                          % (timezone, first_weekday))
        self.calendar = Calendar(timezone=timezone,
                                 first_weekday=first_weekday)

        # Register Stores
        # ===============
        #
        # * Default store is [store] in main config file
        # * Stores are also loaded from main config file from sections with
        #   name [store_*] (not documented feature)

        default = None
        if config.has_section("store"):
            default = dict(config.items("store"))

        if default:
            self._register_store_dict("default", default)

        # Register [store_*] from main config (not documented)
        for section in config.sections():
            if section != "store" and section.startswith("store"):
                name = section[6:]
                self._register_store_dict(name, dict(config.items(section)))

        if config.has_section("browser"):
            self.browser_options = dict(config.items("browser"))
        else:
            self.browser_options = {}

        if config.has_section("main"):
            self.options = dict(config.items("main"))
        else:
            self.options = {}

        # Register Languages
        # ==================
        #

        # Register [language *]
        self.ns_languages = defaultdict(dict)
        for section in config.sections():
            if section.startswith("locale"):
                lang = section[9:]
                # namespace -> path
                for nsname, path in config.items(section):
                    if nsname == "defalt":
                        ns = self.namespace
                    else:
                        (ns, _) = self.namespace.namespace(nsname)
                    ns.add_translation(lang, path)

        # Authorizer
        # ==========

        if config.has_option("workspace", "authorization"):
            auth_type = config.get("workspace", "authorization")
            options = dict(config.items("authorization"))
            options["cubes_root"] = self.root_dir
            self.authorizer = ext.authorizer(auth_type, **options)
        else:
            self.authorizer = None

        # Configure and load models
        # =========================

        # Models are searched in:
        # [model]
        # [models] * <- depreciated!
        # TODO: add this for nicer zero-conf
        # root/model.json
        # root/main.cubesmodel
        # models/*.cubesmodel
        models = []
        # Undepreciated
        if config.has_section("model"):
            if not config.has_option("model", "path"):
                raise ConfigurationError("No model path specified")

            path = config.get("model", "path")
            models.append(("main", path))

        # TODO: Depreciate this too
        if config.has_section("models"):
            models += config.items("models")

        for model, path in models:
            self.logger.debug("Loading model %s" % model)
            self.import_model(path)

    def flush_lookup_cache(self):
        """Flushes the cube lookup cache."""
        self._cubes.clear()
        # TODO: flush also dimensions

    def _get_namespace(self, ref):
        """Returns namespace with ference `ref`"""
        if not ref or ref == "default":
            return self.namespace
        return self.namespace.namespace(ref)[0]

    def add_translation(self, locale, trans, ns="default"):
        """Add translation `trans` for `locale`. `ns` is a namespace. If no
        namespace is specified, then default (global) is used."""

        namespace = self._get_namespace(ns)
        namespace.add_translation(locale, trans)

    def _register_store_dict(self, name, info):
        info = dict(info)
        try:
            type_ = info.pop("type")
        except KeyError:
            try:
                type_ = info.pop("backend")
            except KeyError:
                raise ConfigurationError("Store '%s' has no type specified" % name)
            else:
                self.logger.warn("'backend' is depreciated, use 'type' for "
                                 "store (in %s)." % str(name))

        self.register_store(name, type_, **info)

    def register_default_store(self, type_, **config):
        """Convenience function for registering the default store. For more
        information see `register_store()`"""
        self.register_store("default", type_, **config)

    def register_store(self, name, type_, include_model=True, **config):
        """Adds a store configuration."""

        config = dict(config)

        if name in self.store_infos:
            raise ConfigurationError("Store %s already registered" % name)

        self.store_infos[name] = (type_, config)

        # Model and provider
        # ------------------

        # If store brings a model, then include it...
        if include_model and "model" in config:
            model = config.pop("model")
        else:
            model = None

        # Get related model provider or override it with configuration
        store_factory = ext.store.factory(type_)

        if hasattr(store_factory, "related_model_provider"):
            provider = store_factory.related_model_provider
        else:
            provider = None

        provider = config.pop("model_provider", provider)

        nsname = config.pop("namespace", None)

        if model:
            self.import_model(model, store=name, namespace=nsname,
                              provider=provider)
        elif provider:
            # Import empty model and register the provider
            self.import_model({}, store=name, namespace=nsname,
                              provider=provider)

        self.logger.debug("Registered store '%s'" % name)

    def _store_for_model(self, metadata):
        """Returns a store for model specified in `metadata`. """
        store_name = metadata.get("store")
        if not store_name and "info" in metadata:
            store_name = metadata["info"].get("store")

        store_name = store_name or "default"

        return store_name

    # TODO: this is very confusing process, needs simplification
    # TODO: change this to: add_model_provider(provider, info, store, languages, ns)

    def import_model(self, model=None, provider=None, store=None,
                     translations=None, namespace=None):
        """Registers the `model` in the workspace. `model` can be a
        metadata dictionary, filename, path to a model bundle directory or a
        URL.

        If `namespace` is specified, then the model's objects are stored in
        the namespace of that name.

        `store` is an optional name of data store associated with the model.
        If not specified, then the one from the metadata dictionary will be
        used.

        Model's provider is registered together with loaded metadata. By
        default the objects are registered in default global namespace.

        Note: No actual cubes or dimensions are created at the time of calling
        this method. The creation is deferred until
        :meth:`cubes.Workspace.cube` or :meth:`cubes.Workspace.dimension` is
        called.
        """
        # 1. Metadata
        # -----------
        # Make sure that the metadata is a dictionary
        # 
        # TODO: Use "InlineModelProvider" and "FileBasedModelProvider"

        if store and not isinstance(store, compat.string_type):
            raise ArgumentError("Store should be provided by name "
                                "(as a string).")

        # 1. Model Metadata
        # -----------------
        # Make sure that the metadata is a dictionary
        #
        # TODO: Use "InlineModelProvider" and "FileBasedModelProvider"

        if isinstance(model, compat.string_type):
            self.logger.debug("Importing model from %s. "
                              "Provider: %s Store: %s NS: %s"
                              % (model, provider, store, namespace))
            path = model
            if self.models_dir and not os.path.isabs(path):
                path = os.path.join(self.models_dir, path)
            model = read_model_metadata(path)
        elif isinstance(model, dict):
            self.logger.debug("Importing model from dictionary. "
                              "Provider: %s Store: %s NS: %s"
                              % (provider, store, namespace))
        elif model is None:
            model = {}
        else:
            raise ConfigurationError("Unknown model '%s' "
                                     "(should be a filename or a dictionary)"
                                     % model)

        # 2. Model provider
        # -----------------
        # Create a model provider if name is given. Otherwise assume that the
        # `provider` is a ModelProvider subclass instance

        if isinstance(provider, compat.string_type):
            provider = ext.model_provider(provider, model)

        # TODO: remove this, if provider is external, it should be specified
        if not provider:
            provider_name = model.get("provider", "default")
            provider = ext.model_provider(provider_name, model)

        # 3. Store
        # --------
        # Link the model with store
        store = store or model.get("store")

        if store or (hasattr(provider, "requires_store") \
                        and provider.requires_store()):
            provider.bind(self.get_store(store))

        # 4. Namespace
        # ------------

        if namespace:
            if namespace == "default":
                ns = self.namespace
            elif isinstance(namespace, compat.string_type):
                (ns, _) = self.namespace.namespace(namespace, create=True)
            else:
                ns = namespace
        elif store == "default":
            ns = self.namespace
        else:
            # Namespace with the same name as the store.
            (ns, _) = self.namespace.namespace(store, create=True)

        ns.add_provider(provider)

    def add_slicer(self, name, url, **options):
        """Register a slicer as a model and data provider."""
        self.register_store(name, "slicer", url=url, **options)
        self.import_model({}, provider="slicer", store=name)

    def cube_names(self, identity=None):
        """Return names all available cubes."""
        return [cube["name"] for cube in self.list_cubes()]

    # TODO: this is not loclized!!!
    def list_cubes(self, identity=None):
        """Get a list of metadata for cubes in the workspace. Result is a list
        of dictionaries with keys: `name`, `label`, `category`, `info`.

        The list is fetched from the model providers on the call of this
        method.

        If the workspace has an authorizer, then it is used to authorize the
        cubes for `identity` and only authorized list of cubes is returned.
        """

        all_cubes = self.namespace.list_cubes(recursive=True)

        if self.authorizer:
            by_name = dict((cube["name"], cube) for cube in all_cubes)
            names = [cube["name"] for cube in all_cubes]

            authorized = self.authorizer.authorize(identity, names)
            all_cubes = [by_name[name] for name in authorized]

        return all_cubes

    def cube(self, ref, identity=None, locale=None):
        """Returns a cube with full cube namespace reference `ref` for user
        `identity` and translated to `locale`."""

        if not isinstance(ref, compat.string_type):
            raise TypeError("Reference is not a string, is %s" % type(ref))

        if self.authorizer:
            authorized = self.authorizer.authorize(identity, [ref])
            if not authorized:
                raise NotAuthorized

        # If we have a cached cube, return it
        # See also: flush lookup
        cube_key = (ref, identity, locale)
        if cube_key in self._cubes:
            return self._cubes[cube_key]

        # Find the namespace containing the cube – we will need it for linking
        # later
        (namespace, provider, basename) = self.namespace.find_cube(ref)

        cube = provider.cube(basename, locale=locale, namespace=namespace)
        cube.namespace = namespace
        cube.store = provider.store

        # TODO: cube.ref -> should be ref and cube.name should be basename
        cube.basename = basename
        cube.name = ref

        lookup = namespace.translation_lookup(locale)

        if lookup:
            # TODO: pass lookup instead of jsut first found translation
            context = LocalizationContext(lookup[0])
            trans = context.object_localization("cubes", cube.name)
            cube = cube.localized(trans)

        # Cache the cube
        self._cubes[cube_key] = cube

        return cube

    def dimension(self, name, locale=None, namespace=None, provider=None):
        """Returns a dimension with `name`. Raises `NoSuchDimensionError` when
        no model published the dimension. Raises `RequiresTemplate` error when
        model provider requires a template to be able to provide the
        dimension, but such template is not a public dimension.

        The standard lookup when linking a cube is:

        1. look in the cube's provider
        2. look in the cube's namespace – all providers within that namespace
        3. look in the default (global) namespace
        """

        return find_dimension(name, locale,
                              namespace or self.namespace,
                              provider)

    def _browser_options(self, cube):
        """Returns browser configuration options for `cube`. The options are
        taken from the configuration file and then overriden by cube's
        `browser_options` attribute."""

        options = dict(self.browser_options)
        if cube.browser_options:
            options.update(cube.browser_options)

        return options

    def browser(self, cube, locale=None, identity=None):
        """Returns a browser for `cube`."""

        # TODO: bring back the localization
        # model = self.localized_model(locale)

        if isinstance(cube, compat.string_type):
            cube = self.cube(cube, identity=identity)

        locale = locale or cube.locale

        if isinstance(cube.store, compat.string_type):
            store_name = cube.store or "default"
            store = self.get_store(store_name)
            store_type = self.store_infos[store_name][0]
            store_info = self.store_infos[store_name][1]
        elif cube.store:
            store = cube.store
            store_info = store.options or {}
        else:
            store = self.get_store("default")
            store_info = store.options or {}

        store_type = store.store_type
        if not store_type:
            raise CubesError("Store %s has no store_type set" % store)

        cube_options = self._browser_options(cube)

        # TODO: merge only keys that are relevant to the browser!
        options = dict(store_info)
        options.update(cube_options)

        # TODO: Construct options for the browser from cube's options
        # dictionary and workspece default configuration

        browser_name = cube.browser
        if not browser_name and hasattr(store, "default_browser_name"):
            browser_name = store.default_browser_name
        if not browser_name:
            browser_name = store_type
        if not browser_name:
            raise ConfigurationError("No store specified for cube '%s'" % cube)

        browser = ext.browser(browser_name, cube, store=store,
                              locale=locale, calendar=self.calendar,
                              **options)

        # TODO: remove this once calendar is used in all backends
        browser.calendar = self.calendar

        return browser

    def cube_features(self, cube, identity=None):
        """Returns browser features for `cube`"""
        # TODO: this might be expensive, make it a bit cheaper
        # recycle the feature-providing browser or something. Maybe use class
        # method for that
        return self.browser(cube, identity).features()

    def get_store(self, name=None):
        """Opens a store `name`. If the store is already open, returns the
        existing store."""

        name = name or "default"

        if name in self.stores:
            return self.stores[name]

        try:
            type_, options = self.store_infos[name]
        except KeyError:
            raise ConfigurationError("Unknown store '{}'".format(name))

        # TODO: temporary hack to pass store name and store type
        store = ext.store(type_, store_type=type_, **options)
        self.stores[name] = store
        return store



# -*- coding=utf -*-
from ..model import *
from ..browser import *
from ..stores import Store
from ..providers import ModelProvider
from ..errors import *
from ..logging import get_logger
import json
from .. import compat

DEFAULT_SLICER_URL = "http://localhost:5000"

class _default_opener:
    def __init__(self):
        pass

    def open(self, url, *args, **kwargs):
        return compat.urlopen(url, *args, **kwargs)

class SlicerStore(Store):
    related_model_provider = "slicer"

    __description__ = """
    Uses external Slicer as a store. Aggregation is performed on the remote
    server and results are relayed.
    """
    __options__ = [
        {
            "name": "url",
            "description": "URL of another/external Slicer",
            "type": "string"
        },
        {
            "name": "authentication",
            "description": "Authentication method (pass_parameter or none)",
            "type": "string"
        },
        {
            "name": "auth_identity",
            "description": "Authenticated identity (user name, key, ...)",
            "type": "string"
        },
        {
            "name": "auth_parameter",
            "description": "Name of authentication URL parameter " \
                           "(default: api_key",
            "type": "string"
        },
        {
            "name": "username",
            "description": "HTTP authentication username",
            "type": "string"
        },
        {
            "name": "password",
            "description": "HTTP authentication password",
            "type": "string"
        },
    ]

    def __init__(self, url=None, authentication=None,
                 auth_identity=None, auth_parameter=None,
                 **options):

        super(SlicerStore, self).__init__(**options)

        url = url or DEFAULT_SLICER_URL

        self.url = url
        self.logger = get_logger()

        if authentication and authentication not in ["pass_parameter", "none"]:
            raise ConfigurationError("Unsupported authentication method '%s'"
                                     % authentication)

        self.authentication = authentication
        self.auth_identity = auth_identity
        self.auth_parameter = auth_parameter or "api_key"

        if "username" in options and "password" in options:
            # make a basic auth-enabled opener
            _pmgr = compat.HTTPPasswordMgrWithDefaultRealm()
            _pmgr.add_password(None, self.url, options['username'], options['password'])
            self.opener = compat.build_opener(compat.HTTPBasicAuthHandler(_pmgr))
            self.logger.info("Created slicer opener using basic auth credentials with username %s", options['username'])
        else:
            self.opener = _default_opener()

        # TODO: cube prefix
        # TODO: model mappings as in mixpanel

    def request(self, action, params=None, is_lines=False):
        """
        * `action` – server action (path)
        # `params` – request parameters
        """

        params = dict(params) if params else {}

        if self.authentication == "pass_parameter":
            params[self.auth_parameter] = self.auth_identity

        params_str = compat.urlencode(params)
        request_url = '%s/%s' % (self.url, action)

        if params_str:
            request_url += '?' + params_str

        self.logger.debug("slicer request: %s" % (request_url, ))
        response = self.opener.open(request_url)

        if response.getcode() == 404:
            raise MissingObjectError
        elif response.getcode() != 200:
            raise BackendError("Slicer request error (%s): %s"
                               % (response.getcode(), response.read()))

        if is_lines:
            return _JSONLinesIterator(response)
        else:
            try:
                result = json.loads(response.read())
            except:
                result = {}

            return result

    def cube_request(self, action, cube, params=None, is_lines=False):
        action = "cube/%s/%s" % (cube, action)
        return self.request(action, params, is_lines)


class _JSONLinesIterator(object):
    def __init__(self, stream):
        self.stream = stream

    def __iter__(self):
        for line in self.stream:
            yield json.loads(line)


class SlicerModelProvider(ModelProvider):

    __description__ = """
    Uses external Slicer server as a model provider.
    """

    def requires_store(self):
        return True

    def list_cubes(self):
        return self.store.request('cubes')

    def cube(self, name, locale=None):
        params = {}
        if locale:
            params["lang"] = locale
        try:
            cube_desc = self.store.cube_request("model", name, params)
        except MissingObjectError:
            raise NoSuchCubeError("Unknown cube '%s'" % name, name)

        # create_cube() expects dimensions to be a list of names and linked
        # later, the Slicer returns whole dimension descriptions

        dimensions = cube_desc.pop("dimensions")
        features = cube_desc.pop("features")

        if features:
            # Note: if there are "features" in the browser options, they are
            # eaten here. Is this ok? They should not be there as they should
            # have been processed by the original browser/workspace.
            browser_options = cube_desc.pop("browser_options", {})
            browser_options["features"] = features
            cube_desc["browser_options"] = browser_options

        # Link the cube in-place
        cube = create_cube(cube_desc)
        for dim in dimensions:
            dim = create_dimension(dim)
            cube.add_dimension(dim)

        cube.store = self.store
        return cube

    def dimension(self, name, locale=None, tempaltes=None):
        raise NoSuchDimensionError(name)



# -*- encoding: utf-8 -*-
"""Logical to Physical Mappers"""

from __future__ import absolute_import

import re

from ..errors import ModelError
from ..datastructures import AttributeDict

from .query import to_column


# Note about the future of this module:
#
# Mapper should map the whole schema – mutliple facts and multiple dimensions.
# It should be decoupled from the cube and probably associated with the store
# (or store associated with the mapping)
#

__all__ = (
    "distill_naming",
    "Naming",
    "DEFAULT_KEY_FIELD",

    "Mapper",
    "StarSchemaMapper",
    "DenormalizedMapper",
    "map_base_attributes",
)


DEFAULT_KEY_FIELD = "id"

DEFAULT_FACT_KEY = 'id'
DEFAULT_DIMENSION_KEY = 'id'

# Note: Only keys in this dictionary are allowed in the `naming` dictionary.
# All other keys are ignored.

NAMING_DEFAULTS = {
    "fact_prefix": None,
    "fact_suffix": None,
    "dimension_prefix": None,
    "dimension_suffix": None,
    "dimension_key_prefix": None,
    "dimension_key_suffix": None,

    "denormalized_prefix": None,
    "denormalized_suffix": None,

    "aggregated_prefix": None,
    "aggregated_suffix": None,

    "fact_key": DEFAULT_FACT_KEY,
    "dimension_key": DEFAULT_DIMENSION_KEY,
    "explicit_dimension_primary": False,

    "schema": None,
    "fact_schema": None,
    "dimension_schema": None,
    "aggregate_schema": None,
}


def distill_naming(dictionary):
    """Distill only keys and values related to the naming conventions."""
    d = {key: value for key, value in dictionary.items()
         if key in NAMING_DEFAULTS}

    return Naming(d)


def _match_names(pattern, names):
    """Match names to patterns and return a tuple of matching name with
    extracted value (stripped of suffix/prefix)."""

    result = []

    for name in names:
        match = pattern.match(name)
        if match:
            result.append((name, match.group("name")))

    return result


class Naming(AttributeDict):
    """Naming conventions for SQL tables. Naming properties can be accessed as
    a dictionary keys or as direct attributes. The naming properties are:

    * `fact_prefix` – prefix for fact tables
    * `fact_suffix` – suffix for fact tables
    * `dimension_prefix` – prefix for dimension tables
    * `dimension_suffix` – suffix for dimension tables
    * `dimension_key_prefix` – prefix for dimension foreign keys
    * `dimension_key_suffix` – suffix for dimension foreign keys
    * `fact_key` – name of fact table primary key (defaults to ``id`` if not
      specified)
    * `dimension_key` – name of dimension table primary key (defaults to
      ``id`` if not specified)
    * `explicit_dimension_primary` – whether the primary key of dimension
      table contains dimension name explicitly.

    If the `explicit_dimension_primary` is `True`, then all dimension tables
    are expected to have the primary key in the same format as foreign
    dimension keys. For example if the foreign dimension keys are
    ``customer_key`` then primary key of customer dimension table is also
    ``customer_key`` as oposed to just ``key``. The `dimension_key` naming
    property is ignored.


    Additional information that can be used by the mapper:

    * `schema` – default schema
    * `fact_schema` – schema where all fact tables are stored
    * `dimension_schema` – schema where dimension tables are stored

    Recommended values: `fact_prefix` = ``ft_``, `dimension_prefix` =
    ``dm_``, `explicit_dimension_primary` = ``True``.
    """

    def __init__(self, *args, **kwargs):
        """Creates a `Naming` object instance from a dictionary. If `fact_key`
        or `dimension_key` are not specified, then they are set to ``id`` by
        default."""

        super(Naming, self).__init__(*args, **kwargs)

        # Set the defaults
        for key, value in NAMING_DEFAULTS.items():
            if key not in self:
                self[key] = value

        pat = re.compile("^{}(?P<name>.*){}$"
                         .format(self.dimension_prefix or "", self.dimension_suffix or ""))
        self["dim_name_pattern"] = pat

        pat = re.compile("^{}(?P<name>.*){}$"
                         .format(self.fact_prefix or "", self.fact_suffix or ""))
        self["fact_name_pattern"] = pat

        pat = re.compile("^{}(?P<name>.*){}$"
                         .format(self.dimension_key_prefix or "", self.dimension_key_suffix or ""))
        self["dim_key_pattern"] = pat

    def dimension_table_name(self, name):
        """Constructs a physical dimension table name for dimension `name`"""

        table_name = "{}{}{}".format(self.dimension_prefix or "",
                                     name,
                                     self.dimension_suffix or "")
        return table_name

    def fact_table_name(self, name):
        """Constructs a physical fact table name for fact/cube `name`"""

        table_name = "{}{}{}".format(self.fact_prefix or "",
                                     name,
                                     self.fact_suffix or "")
        return table_name

    def denormalized_table_name(self, name):
        """Constructs a physical fact table name for fact/cube `name`"""

        table_name = "{}{}{}".format(self.denormalized_prefix or "",
                                     name,
                                     self.denormalized_suffix or "")
        return table_name

    # TODO: require list of dimensions here
    def aggregated_table_name(self, name):
        """Constructs a physical fact table name for fact/cube `name`"""

        table_name = "{}{}{}".format(self.aggregated_prefix or "",
                                     name,
                                     self.aggregated_suffix or "")
        return table_name

    def dimension_primary_key(self, name):
        """Constructs a dimension primary key name for dimension `name`"""

        if self.explicit_dimension_primary:
            key = "{}{}{}".format(self.dimension_key_prefix or "",
                                  name,
                                  self.dimension_key_suffix or "")
            return key
        else:
            return self.dimension_key

    def dimension_keys(self, keys):
        """Return a list of tuples (`key`, `dimension`) for every key in
        `keys` that matches dimension key naming. Useful when trying to
        identify dimensions and their foreign keys in a fact table that
        follows the naming convetion."""

        return _match_names(self.dim_key_pattern, keys)

    def dimensions(self, table_names):
        """Return a list of tuples (`table`, `dimension`) for all tables that
        match dimension naming scheme. Usefult when trying to identify
        dimension tables in a database that follow the naming convention."""

        return _match_names(self.dim_name_pattern, table_names)

    def facts(self, table_names):
        """Return a list of tuples (`table`, `fact`) for all tables that
        match fact table naming scheme. Useful when trying to identify fact
        tables in a database that follow the naming convention."""

        return _match_names(self.fact_name_pattern, table_names)


class Mapper(object):
    """A dictionary-like object that provides physical column references for
    cube attributes. Does implicit mapping of an attribute.

    .. versionchanged:: 1.1
    """

    def __init__(self, cube, naming, locale=None):
        """Creates a mapping for `cube` using `naming` conventions within
        optional `locale`. `naming` has to be a :class:`cubes.Naming`
        object."""
        self.naming = naming
        self.locale = locale
        self.mappings = cube.mappings or {}
        self.fact_name = cube.fact or naming.fact_table_name(cube.name)


    def __getitem__(self, attribute):
        """Returns implicit physical column reference for `attribute`, which
        should be an instance of :class:`cubes.model.Attribute`. If there is
        no dimension specified in attribute, then fact table is assumed. The
        returned reference has attributes `schema`, `table`, `column`,
        `extract`.  """

        column_name = attribute.name

        if attribute.is_localizable():
            locale = self.locale if self.locale in attribute.locales \
                                else attribute.locales[0]

            column_name = "{}_{}".format(column_name, locale)

        schema, table = self.attribute_table(attribute)

        return to_column((schema, table, column_name))

    def attribute_table(self, attribute):
        """Return a tuple (schema, table) for attribute."""

        dimension = attribute.dimension

        if dimension:
            schema = self.naming.dimension_schema or self.naming.schema
            if dimension.is_flat and not dimension.has_details:
                table = self.fact_name
            else:
                table = self.naming.dimension_table_name(dimension)

        else:
            table = self.fact_name
            schema = self.naming.schema

        return (schema, table)


class DenormalizedMapper(Mapper):
    def __getitem__(self, attribute):
        if attribute.expression:
            raise ModelError("Attribute '{}' has an expression, it can not "
                             "have a direct physical representation"
                             .format(attribute.name))

        return super(DenormalizedMapper, self).__getitem__(attribute)


class StarSchemaMapper(Mapper):
    def __getitem__(self, attribute):
        """Find physical reference for a star schema as follows:

        1. if there is mapping for `dimension.attribute`, use the mapping
        2. if there is no mapping or no mapping was found, then use table
        `dimension` or fact table, if attribute does not belong to a
        dimension and column `attribute`

        If table prefixes and suffixes are used, then they are
        prepended/appended to the table tame in the implicit mapping.

        If localization is requested and the attribute is localizable, then
        suffix in the form `_LOCALE` where `LOCALE` is the locale name will be
        added to search for mapping or for implicit attribute creation such as
        `name_sk` for attribute `name` and locale `sk`.
        """

        if attribute.expression:
            raise ModelError("Attribute '{}' has an expression, it can not "
                             "have a direct physical representation"
                             .format(attribute.name))

        # Fix locale: if attribute is not localized, use none, if it is
        # localized, then use specified if exists otherwise use default
        # locale of the attribute (first one specified in the list)

        if attribute.is_localizable():
            locale = self.locale if self.locale in attribute.locales \
                                else attribute.locales[0]
        else:
            locale = None

        logical = attribute.localized_ref(locale)

        physical = self.mappings.get(logical)

        if physical:
            # TODO: Should we not get defaults here somehow?
            column = to_column(physical)
            return column

        # No mappings exist or no mapping was found - we are going to create
        # default physical reference
        return super(StarSchemaMapper, self).__getitem__(attribute)


def map_base_attributes(cube, mapper_class, naming, locale=None):
    """Map all base attributes of `cube` using mapping function `mapper`.
    `naming` is a naming convention object. Returns a tuple (`fact_name`,
    `mapping`) where `fact_name` is a fact table name and `mapping` is a
    dictionary of attribute references and their physical column
    references."""

    base = [attr for attr in cube.all_attributes if attr.is_base]

    mapper = mapper_class(cube, naming, locale)
    mapped = {attr.ref:mapper[attr] for attr in base}

    return (mapper.fact_name, mapped)




# -*- coding: utf-8 -*-
"""Logical model model providers."""
import copy

from .errors import ModelError, TemplateRequired, CubesError, BackendError
from .errors import NoSuchDimensionError, NoSuchCubeError
from .localization import LocalizationContext
from .model import Cube, Dimension

__all__ = [
    "ModelProvider",
    "StaticModelProvider",
    "link_cube",
]


# Proposed Provider API:
#     Provider.cube() – in abstract class
#     Provider.provide_cube() – in concrete class, providers Cube object that
#         might be modified later
#     Provider.provide_dimension()
#     Provider.link_cube(cube,locale)
#     Provider.find_dimension(cube, locale)
#
# Provider is bound to namespace

# TODO: add tests
# TODO: needs to be reviewed
def link_cube(cube, locale, provider=None, namespace=None,
              ignore_missing=False):
    """Links dimensions to the `cube` in the `context` object. The `context`
    object should implement a function `dimension(name, locale, namespace,
    provider)`. Modifies cube in place, returns the cube.
    """
    # TODO: change this to: link_cube(cube, locale, namespace, provider)

    # Assumption: empty cube

    linked = set()

    for dim_name in cube.dimension_links.keys():
        if dim_name in linked:
            raise ModelError("Dimension '{}' linked twice"
                             .format(dim_name))

        try:
            dim = find_dimension(dim_name, locale,
                                 provider=provider,
                                 namespace=namespace)

        except TemplateRequired as e:
            raise ModelError("Dimension template '%s' missing" % dim_name)

        if not dim and not ignore_missing:
            raise CubesError("Dimension '{}' not found.".format(dim_name))

        cube.link_dimension(dim)

    return cube


# TODO: add tests
def find_dimension(name, locale=None, provider=None, namespace=None):
    """Returns a localized dimension with `name`. Raises
    `NoSuchDimensionError` when no model published the dimension. Raises
    `RequiresTemplate` error when model provider requires a template to be
    able to provide the dimension, but such template is not a public
    dimension.

    The standard lookup when linking a cube is:

    1. look in the provider
    2. look in the namespace – all providers within that namespace
    """

    # Collected dimensions – to be used as templates
    templates = {}

    # Assumption: all dimensions that are to be used as templates should
    # be public dimensions. If it is a private dimension, then the
    # provider should handle the case by itself.
    missing = [name]

    while missing:
        dimension = None

        name = missing.pop()

        # First give a chance to provider, then to namespace
        dimension = None
        required_template = None

        try:
            dimension = _lookup_dimension(name, templates,
                                          namespace, provider)
        except TemplateRequired as e:
            required_template = e.template

        if required_template in templates:
            raise BackendError("Some model provider didn't make use of "
                               "dimension template '%s' for '%s'"
                               % (required_template, name))

        if required_template:
            missing.append(name)
            if required_template in missing:
                raise ModelError("Dimension templates cycle in '%s'" %
                                 required_template)
            missing.append(required_template)

        # Store the created dimension to be used as template
        if dimension:
            templates[name] = dimension

    if namespace:
        lookup = namespace.translation_lookup(locale)

        if lookup:
            # TODO: pass lookup instead of jsut first found translation
            context = LocalizationContext(lookup[0])
            trans = context.object_localization("dimensions", "inner")
            dimension = dimension.localized(trans)

    return dimension


# TODO: add tests
def _lookup_dimension(name, templates, namespace, provider):
    """Look-up a dimension `name` in `provider` and then in `namespace`.

    `templates` is a dictionary with already instantiated dimensions that
    can be used as templates.
    """

    dimension = None

    # 1. look in the povider
    if provider:
        try:
            dimension = provider.dimension(name, templates=templates)
        except NoSuchDimensionError:
            pass
        else:
            return dimension

    # 2. Look in the namespace
    if namespace:
        return namespace.dimension(name, templates=templates)

    raise NoSuchDimensionError("Dimension '%s' not found" % name,
                               name=name)


class ModelProvider(object):
    """Abstract class – factory for model object. Currently empty and used
    only to find other model providers."""

    # TODO: Don't get metadata, but arbitrary arguments.
    def __init__(self, metadata=None):
        """Base class for model providers. Initializes a model provider and
        sets `metadata` – a model metadata dictionary.

        Subclasses should call this method at the beginning of the custom
        `__init__()`.

        If a model provider subclass has a metadata that should be pre-pended
        to the user-provided metadta, it should return it in
        `default_metadata()`.

        Subclasses should implement at least: :meth:`cubes.ModelProvider.cube`,
        :meth:`cubes.ModelProvider.dimension` and
        :meth:`cubes.ModelProvider.list_cubes` methods.
        """

        self.store = None

        # Get provider's defaults and pre-pend it to the user provided
        # metadtata.
        defaults = self.default_metadata()
        self.metadata = self._merge_metadata(defaults, metadata)

        # TODO: check for duplicates
        self.dimensions_metadata = {}
        for dim in self.metadata.get("dimensions", []):
            self.dimensions_metadata[dim["name"]] = dim

        self.cubes_metadata = {}
        for cube in self.metadata.get("cubes", []):
            self.cubes_metadata[cube["name"]] = cube

        # TODO: decide which one to use
        self.options = self.metadata.get("options", {})
        self.options.update(self.metadata.get("browser_options", {}))

    def _merge_metadata(self, metadata, other):
        """See `default_metadata()` for more information."""

        metadata = dict(metadata)
        other = dict(other)

        cubes = metadata.pop("cubes", []) + other.pop("cubes", [])
        if cubes:
            metadata["cubes"] = cubes

        dims = metadata.pop("dimensions", []) + other.pop("dimensions", [])
        if dims:
            metadata["dimensions"] = dims

        joins = metadata.pop("joins", []) + other.pop("joins",[])
        if joins:
            metadata["joins"] = joins

        mappings = metadata.pop("mappings", {})
        mappings.update(other.pop("mappings", {}))
        if mappings:
            metadata["mappings"] = mappings

        metadata.update(other)

        return metadata

    def default_metadata(self, metadata=None):
        """Returns metadata that are prepended to the provided model metadata.
        `metadata` is user-provided metadata and might be used to decide what
        kind of default metadata are returned.

        The metadata are merged as follows:

        * cube lists are concatenated (no duplicity checking)
        * dimension lists are concatenated (no duplicity checking)
        * joins are concatenated
        * default mappings are updated with the model's mappings

        Default implementation returns empty metadata.
        """

        return {}

    # TODO: remove this in favor of provider configuration: store=
    def requires_store(self):
        """Return `True` if the provider requires a store. Subclasses might
        override this method. Default implementation returns `False`"""
        return False

    # TODO: bind this automatically on provider configuration: store (see
    # requires_store() function)
    def bind(self, store):
        """Set's the provider's `store`. """

        self.store = store
        self.initialize_from_store()

    def initialize_from_store(self):
        """This method is called after the provider's `store` was set.
        Override this method if you would like to perform post-initialization
        from the store."""
        pass

    def cube_options(self, cube_name):
        """Returns an options dictionary for cube `name`. The options
        dictoinary is merged model `options` metadata with cube's `options`
        metadata if exists. Cube overrides model's global (default)
        options."""

        options = dict(self.options)
        if cube_name in self.cubes_metadata:
            cube = self.cubes_metadata[cube_name]
            # TODO: decide which one to use
            options.update(cube.get("options", {}))
            options.update(cube.get("browser_options", {}))

        return options

    def dimension_metadata(self, name, locale=None):
        """Returns a metadata dictionary for dimension `name` and optional
        `locale`.

        Subclasses should override this method and call the super if they
        would like to merge metadata provided in a model file."""

        try:
            return self.dimensions_metadata[name]
        except KeyError:
            raise NoSuchDimensionError("No such dimension '%s'" % name, name)

    def cube_metadata(self, name, locale=None):
        """Returns a cube metadata by combining model's global metadata and
        cube's metadata. Merged metadata dictionaries: `browser_options`,
        `mappings`, `joins`.

        Subclasses should override this method and call the super if they
        would like to merge metadata provided in a model file.

        .. note:

            If provider is caching a cube metadata, it should store a cache
            for localized version of the cube metadata.
        """

        if name in self.cubes_metadata:
            metadata = dict(self.cubes_metadata[name])
        else:
            raise NoSuchCubeError("No such cube '%s'" % name, name)

        # merge browser_options
        browser_options = self.metadata.get('browser_options', {})
        if metadata.get('browser_options'):
            browser_options.update(metadata.get('browser_options'))
        metadata['browser_options'] = browser_options

        # Merge model and cube mappings
        #
        model_mappings = self.metadata.get("mappings")
        cube_mappings = metadata.pop("mappings", {})

        if model_mappings:
            mappings = copy.deepcopy(model_mappings)
            mappings.update(cube_mappings)
        else:
            mappings = cube_mappings

        metadata["mappings"] = mappings

        # Merge model and cube joins
        #
        model_joins = self.metadata.get("joins", [])
        cube_joins = metadata.pop("joins", [])

        # model joins, if present, should be merged with cube's overrides.
        # joins are matched by the "name" key.
        if cube_joins and model_joins:
            model_join_map = {}
            for join in model_joins:
                try:
                    jname = join['name']
                except KeyError:
                    raise ModelError("Missing required 'name' key in "
                                     "model-level joins.")

                if jname in model_join_map:
                    raise ModelError("Duplicate model-level join 'name': %s" %
                                     jname)

                model_join_map[jname] = copy.deepcopy(join)

            # Merge cube's joins with model joins by their names.
            merged_joins = []

            for join in cube_joins:
                name = join.get('name')
                if name and name in model_join_map:
                    model_join = dict(model_join_map[name])
                else:
                    model_join = {}

                model_join.update(join)
                merged_joins.append(model_join)
        else:
            merged_joins = cube_joins

        # Validate joins:
        for join in merged_joins:
            if "master" not in join:
                raise ModelError("No master in join for cube '%s' "
                                 "(join name: %s)" % (name, join.get("name")))
            if "detail" not in join:
                raise ModelError("No detail in join for cube '%s' "
                                 "(join name: %s)" % (name, join.get("name")))

        metadata["joins"] = merged_joins

        return metadata

    def list_cubes(self):
        """Get a list of metadata for cubes in the workspace. Result is a list
        of dictionaries with keys: `name`, `label`, `category`, `info`.

        The list is fetched from the model providers on the call of this
        method.

        Subclassees should implement this method.
        """
        raise NotImplementedError("Subclasses should implement list_cubes()")

    def has_cube(self, name):
        """Returns `True` if the provider has cube `name`. Otherwise returns
        `False`."""

        return name in self.cubes_metadata

    def cube(self, name, locale=None, namespace=None):
        """Returns a cube with `name` provided by the receiver. If receiver
        does not have the cube `NoSuchCube` exception is raised.

        Note: The returned cube will not have the dimensions assigned.
        It is up to the caller's responsibility to assign appropriate
        dimensions based on the cube's `dimension_links`.

        Subclasses of `ModelProvider` might override this method if they would
        like to create the `Cube` object directly.

        .. note:

            If provider is caching a cube, it should store a cache for
            localized version of the cube.
        """

        metadata = self.cube_metadata(name, locale)
        cube = Cube.from_metadata(metadata)
        link_cube(cube, locale, provider=self, namespace=namespace)

        return cube

    def dimension(self, name, templates=[], locale=None):
        """Returns a dimension with `name` provided by the receiver.
        `dimensions` is a dictionary of dimension objects where the receiver
        can look for templates. If the dimension requires a template and the
        template is missing, the subclasses should raise
        `TemplateRequired(template)` error with a template name as an
        argument.

        If the receiver does not provide the dimension `NoSuchDimension`
        exception is raised.
        """
        metadata = self.dimension_metadata(name, locale)
        return Dimension.from_metadata(metadata, templates=templates)


# TODO: make this FileModelProvider
class StaticModelProvider(ModelProvider):

    __extension_aliases__ = ["default"]

    def __init__(self, *args, **kwargs):
        super(StaticModelProvider, self).__init__(*args, **kwargs)
        # Initialization code goes here...

    def list_cubes(self):
        """Returns a list of cubes from the metadata."""
        cubes = []

        for cube in self.metadata.get("cubes", []):
            info = {
                "name": cube["name"],
                "label": cube.get("label", cube["name"]),
                "category": (cube.get("category") or cube.get("info", {}).get("category")),
                "info": cube.get("info", {})
            }
            cubes.append(info)

        return cubes




# -*- coding: utf-8 -*-
"""Exceptions used in Cubes.

The base exception calss is :class:`.CubesError`."""

from collections import OrderedDict

class CubesError(Exception):
    """Generic error class."""

class InconsistencyError(CubesError):
    """Raised when something bad happened in cubes – very likely an edge
    case that is not handled properly.

    It is very unlikely that the user might fix this error by changing his/her
    input.
    """

class UserError(CubesError):
    """Superclass for all errors caused by the cubes and slicer users. Error
    messages from this error might be safely passed to the front-end. Do not
    include any information that you would not like to be public.

    Users can fix the error."""
    error_type = "unknown_user_error"

class InternalError(CubesError):
    """Superclass for all errors that happened on the server side:
    configuration issues, connection problems, model inconsistencies...

    If you handle this exception, don't display content of this error to the
    clients (such as over the web), as it might contain information about the
    server configuration, database or other internals.
    """
    error_type = "internal_error"

class ConfigurationError(InternalError):
    """Raised when there is a problem with workspace configuration assumed."""

class BackendError(InternalError):
    """Raised by a backend. Should be handled separately, for example: should
    not be passed to the client from the server due to possible internal
    schema exposure.
    """

class WorkspaceError(InternalError):
    """Backend Workspace related exception."""

class BrowserError(InternalError):
    """AggregationBrowser related exception."""
    pass

class StoreError(InternalError):
    """AggregationBrowser related exception."""
    pass

class ModelError(InternalError):
    """Model related exception."""

class ExpressionError(ModelError):
    """Expression related exception such as unknown attribute or cirular
    attribute reference"""

# TODO: necessary? or rename to PhysicalModelError
class MappingError(ModelError):
    """Raised when there are issues by mapping from logical model to physical
    database schema. """


# TODO: change all instances to ModelError
class ModelInconsistencyError(ModelError):
    """Raised when there is incosistency in model structure."""

class MissingObjectError(UserError):
    error_type = "missing_object"
    object_type = None

    def __init__(self, message=None, name=None):
        self.message = message
        self.name = name

    def __str__(self):
        return self.message or self.name

    def to_dict(self):
        d = OrderedDict()
        d["object"] = self.name
        d["message"] = self.message
        if self.object_type:
            d["object_type"] = self.object_type

        return d

class NoSuchDimensionError(MissingObjectError):
    """Raised when an unknown dimension is requested."""
    object_type = "dimension"

class NoSuchCubeError(MissingObjectError):
    """Raised when an unknown cube is requested."""
    object_type = "cube"

class NoSuchAttributeError(UserError):
    """Raised when an unknown attribute, measure or detail requested."""
    object_type = "attribute"

class ArgumentError(UserError):
    """Raised when an invalid or conflicting function argument is supplied.
    """

class HierarchyError(UserError):
    """Raised when attemt to get level deeper than deepest level in a
    hierarchy"""
    error_type = "hierarchy"


# Helper exceptions
# =================
#
# Not quite errors, but used for signalling
#
class TemplateRequired(ModelError):
    """Raised by a model provider which can provide a dimension, but requires
    a template. Signals to the caller that the creation of a dimension should
    be retried when the template is available."""

    def __init__(self, template):
        self.template = template
    def __str__(self):
        return self.template




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from logging import getLogger, Formatter, StreamHandler, FileHandler

__all__ = [
    "get_logger",
    "create_logger",
]

DEFAULT_LOGGER_NAME = "cubes"
DEFAULT_FORMAT = "%(asctime)s %(levelname)s %(message)s"
logger = None

# TODO: make name first
def get_logger(path=None, format_=None, name=None):
    """Get brewery default logger"""
    global logger

    if logger:
        return logger
    else:
        return create_logger(path, format_, name)

def create_logger(path=None, format_=None, name=None):
    """Create a default logger"""
    global logger
    logger = getLogger(name or DEFAULT_LOGGER_NAME)
    logger.propagate = False

    if not logger.handlers:
        formatter = Formatter(fmt=format_ or DEFAULT_FORMAT)

        if path:
            # create a logger which logs to a file
            handler = FileHandler(path)
        else:
            # create a default logger
            handler = StreamHandler()

        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger




# -*- encoding: utf-8 -*-

from __future__ import absolute_import
from .blueprint import slicer
from flask import Flask
import shlex
import os

from .utils import *
from .. import compat
from ..logging import get_logger

__all__ = (
    "create_server",
    "run_server"
)

# Server Instantiation and Running
# ================================

def read_slicer_config(config):
    if not config:
        return compat.ConfigParser()
    elif isinstance(config, compat.string_type):
        try:
            path = config
            config = compat.ConfigParser()
            config.read(path)
        except Exception as e:
            raise Exception("Unable to load configuration: %s" % e)
    return config

def create_server(config=None, **_options):
    """Returns a Flask server application. `config` is a path to a
    ``slicer.ini`` file with Cubes workspace and server configuration."""

    # Load extensions
    if config.has_option("server", "modules"):
        modules = shlex.split(config.get("server", "modules"))
        for module in modules:
            e = __import__(module)

    app = Flask(__name__.rsplit('.', 1)[0])
    # FIXME: read note about _options in Workspace. Only for internal use as a
    # temporary quick fix.
    app.register_blueprint(slicer, config=config, **_options)

    return app

def run_server(config, debug=False, app=None):
    """Run OLAP server with configuration specified in `config`"""

    config = read_slicer_config(config)

    logger = get_logger()

    if config.has_option("server", "debug"):
        if debug is False and config.getboolean("server", "debug"):
            debug = True

    if debug:
        logger.warning('Server running under DEBUG, so logging level set to DEBUG.')
        import logging
        logger.setLevel(logging.DEBUG)

    if app is None:
        app = create_server(config)

    if config.has_option("server", "host"):
        host = config.get("server", "host")
    else:
        host = "localhost"

    if config.has_option("server", "port"):
        port = config.getint("server", "port")
    else:
        port = 5000

    if config.has_option("server", "reload"):
        use_reloader = config.getboolean("server", "reload")
    else:
        use_reloader = False

    if config.has_option('server', 'processes'):
        processes = config.getint('server', 'processes')
    else:
        processes = 1

    if config.has_option("server", "pid_file"):
        path = config.get("server", "pid_file")
        try:
            with open(path, "w") as f:
                f.write(str(os.getpid()))
        except IOError as e:
            logger.error("Unable to write PID file '%s'. Check the "
                         "directory existence or permissions." % path)
            raise

    app.run(host, port, debug=debug, processes=processes,
            use_reloader=use_reloader)

